﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace XnaMapGenerator3D.Models
{
    public class Triangle
    {
        private VertexPositionNormalColor[] vertices = new VertexPositionNormalColor[3];

        public Triangle(Center center, Corner second, Corner third)
        {
            if (Area(center.Point, second.Point, third.Point) > 0)
            {
                vertices[0] = new VertexPositionNormalColor(
                                new Vector3(center.Point.X, center.Point.Y, center.Point.Z),
                                center.Normal,
                                center.Biome.Color
                                );

                vertices[1] = new VertexPositionNormalColor(
                                    new Vector3(second.Point.X, second.Point.Y, second.Point.Z),
                                    second.Normal,
                                    (center.Water) ? center.Biome.Color : second.Biome.Color);

                vertices[2] = new VertexPositionNormalColor(
                                    new Vector3(third.Point.X, third.Point.Y, third.Point.Z),
                                    third.Normal,
                                    (center.Water) ? center.Biome.Color : third.Biome.Color);
            }
            else
            {
                vertices[0] = new VertexPositionNormalColor(
                            new Vector3(center.Point.X, center.Point.Y, center.Point.Z),
                            center.Normal,
                            center.Biome.Color);

                vertices[1] = new VertexPositionNormalColor(
                                    new Vector3(third.Point.X, third.Point.Y, third.Point.Z),
                                    third.Normal,
                                    (center.Water) ? center.Biome.Color : third.Biome.Color);

                vertices[2] = new VertexPositionNormalColor(
                                    new Vector3(second.Point.X, second.Point.Y, second.Point.Z),
                                    second.Normal,
                                    (center.Water) ? center.Biome.Color : second.Biome.Color);
            }
        }

        public float Area(Vector3 A, Vector3 B, Vector3 C)
        {
            return ((A.X - C.X) * (B.Z - C.Z) - (A.Z - C.Z) * (B.X - C.X)) / 2;
        }

        public void Draw(GraphicsDevice device)
        {
            device.DrawUserPrimitives(PrimitiveType.TriangleList, vertices, 0, 1);    
        }
    }

    public class Polygon
    {
        private List<Triangle> tris = new List<Triangle>();

        public Polygon(HashSet<Corner> Corners, Center center)
        {
            var cnr = new Corner[Corners.Count + 1];

            for (int i = 0; i < cnr.Count() - 1; i++)
            {
                cnr[i] = Corners.ElementAt(i);
            }

            cnr[Corners.Count] = cnr.First();

            for (int i = cnr.Count() - 1; i >= 1; i--)
            {
                tris.Add(new Triangle(center, cnr[i], cnr[i - 1]));
            }
        }

        private Vector3 CalculateNormal(Vector3 newMain, Vector3 p1, Vector3 p2)
        {
            var v1 = new Vector3(newMain.X - p1.X, newMain.Y - p1.Y, newMain.Z - p1.Z);
            var v2 = new Vector3(p1.X - p2.X, p1.Y - p2.Y, p1.Z - p2.Z);
            return Vector3.Cross(v1, v2);
        }

        private Vector3 CalculateNormal2(Vector3 newMain, Vector3 p1, Vector3 p2)
        {
            var v1 = new Vector3(newMain.X - p2.X, newMain.Y - p2.Y, newMain.Z - p2.Z);
            var v2 = new Vector3(p2.X - p1.X, p2.Y - p1.Y, p2.Z - p1.Z);
            return Vector3.Cross(v1, v2);
        }

        public void Draw(GraphicsDevice device)
        {
            foreach (Triangle triangle in tris)
            {
                triangle.Draw(device);
            }
        }
    }
}
